import pytest

from tests.fixtures.route import *
from typer_router import Router


@pytest.fixture
def app1_router(
    app1_command1_route,
    app1_command1_1_route,
    app1_command1_2_route,
    app1_command2_1_1_route,
) -> Router:
    return Router(
        routes=[
            app1_command1_route,
            app1_command1_1_route,
            app1_command1_2_route,
            app1_command2_1_1_route,
        ],
        app_import_path=APP_1_IMPORT_PATH,
    )



import importlib
from typing import TYPE_CHECKING, Dict, Set

import typer

from typer_router.exc import NoParentRouteException

if TYPE_CHECKING:
    from typer_router.route import Route
    from typer_router.router import Router

TyperDict = Dict[str, typer.Typer]


def create_typer_app_from_router(router: "Router", **typer_kwargs) -> typer.Typer:
    app = typer.Typer(**typer_kwargs)
    nested_apps: TyperDict = {}

    for route in router.routes:
        # Handle the parents of the leaf nodes, as those will just be container typers
        try:
            parent = route.parent
        except NoParentRouteException:
            parent = None

        if parent:
            container_typer = nested_apps.get(parent.import_path)
            if container_typer is None:
                container_typer = typer.Typer(name=parent.name)
                nested_apps[parent.import_path] = container_typer
            # Connect the leaf node to the container
            add_route_to_typer_app(route, router, container_typer)
        else:
            # This is a root command, add directly to main typer
            add_route_to_typer_app(route, router, app)

    # Now create the nested container typers to fill in the gaps.
    # For example, if we have the following routes:
    #  - a.b.c.d
    #  - a.b.f.g
    #  - a.b.h.i.j
    # Then c, d, f, g, i, and j have already been created but b and h still need to be created.
    for route in router.routes:
        for sub_route in route.subroutes:
            path = sub_route.import_path
            if path not in nested_apps:
                container_typer = typer.Typer(name=sub_route.name)
                nested_apps[path] = container_typer
                # app.add_typer(container_typer, name=sub_route.name)

    # Now one last pass to connect all the typers together.
    # The leaf nodes and their containers are already connected
    # But we need to connect any intermediate containers and also connect
    # the top level containers to the main app.
    # For example, if we have the following routes:
    #  - a.b.c.d
    #  - a.b.f.g
    #  - a.b.h.i.j
    #  - a.k
    # We now need to connect b to a, k to a, and h to b.
    # This will make more sense after reading the prior block of comments
    already_connected: Set[str] = {route.import_path for route in router.routes}
    # Start from connecting to the main app
    for route in router.routes:
        # Need to create from leaf inwards so that commands will exist when connecting to main app
        for sub_route in reversed(route.subroutes):
            # E.g. for sub_route:
            # a.b.c.d we will get a.b.c, a.b, a
            # a.b.h.i.j we will get a.b.h.i, a.b.h, a.b, a
            # a.k we will get a
            if sub_route.import_path in already_connected:
                continue
            app_to_connect = nested_apps[sub_route.import_path]
            try:
                parent_app = nested_apps[sub_route.parent.import_path]
            except NoParentRouteException:
                parent_app = app
            parent_app.add_typer(app_to_connect, name=sub_route.name)
            already_connected.add(sub_route.import_path)

    return app


def add_route_to_typer_app(route: "Route", router: "Router", app: typer.Typer):
    # Load the python file from the route's import path
    # and get the route's function from the file
    full_import_path = router.full_import_path_for(route)
    module = importlib.import_module(full_import_path)
    func = getattr(module, route.function_name)

    app.command(route.name)(func)



import typer

from tests.e2e.cli_stub import run_cli


def test_app_1_command_1_prints_a_b(app1_typer: typer.Typer):
    result = run_cli(app1_typer, "command1 --b 2 1")
    assert result.exit_code == 0
    assert result.stdout.strip() == "a=1, b=2"


def test_app_1_command_1_1_prints_e_f(app1_typer: typer.Typer):
    result = run_cli(app1_typer, "container1 command1_1 --f 2 1")
    assert result.exit_code == 0
    assert result.stdout.strip() == "e=1, f=2"


def test_app_1_command_1_2_prints_g_h(app1_typer: typer.Typer):
    result = run_cli(app1_typer, "container1 command1_2 --h 2 1")
    assert result.exit_code == 0
    assert result.stdout.strip() == "g=1, h=2"


def test_app_1_command_2_1_1_prints_c_d(app1_typer: typer.Typer):
    result = run_cli(app1_typer, "container2 container2_1 command2_1_1 --d 2 1")
    assert result.exit_code == 0
    assert result.stdout.strip() == "c=1, d=2"



import shlex
from enum import Enum

import typer
from click.testing import Result
from typer.testing import CliRunner

from tests.e2e import ext_click

runner = CliRunner()


class CLIRunnerException(Exception):
    pass


class ExceptionHandling(str, Enum):
    RAISE = "raise"
    IGNORE = "ignore"


def run_cli(
    app: typer.Typer,
    command: str,
    exception_handling: ExceptionHandling = ExceptionHandling.RAISE,
) -> Result:
    split_command = shlex.split(command)
    result = runner.invoke(app, split_command)
    if exception_handling == ExceptionHandling.RAISE and result.exit_code != 0:
        output = ext_click.result_to_message(result)
        full_command = shlex.join(["terminhtml", command])
        raise CLIRunnerException(
            f"{full_command} exited with code {result.exit_code}.\n{output}"
        )
    return result



from typing import List

from click.testing import Result


def result_to_message(result: Result) -> str:
    output_parts: List[str] = []
    if result.stdout:
        output_parts.append(f"stdout:\n{result.stdout}")
    try:
        output_parts.append(f"sterr:\n{result.stderr}")
    except ValueError as e:
        if "stderr not separately captured" in str(e):
            pass
        else:
            raise e
    if result.exception:
        output_parts.append(f"exception:\n{result.exception}")
    return "\n".join(output_parts)



import importlib
from pathlib import Path
from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from typer_router.route import Route


def routes_from_app_import_path(app_import_path: str) -> List["Route"]:
    """Create routes from an app import path."""
    from typer_router.route import Route

    app_path = _app_folder_from_import_path(app_import_path)
    # Iterate through nested directories and files
    routes: List[Route] = []
    for path in app_path.rglob("*"):
        if path.is_dir():
            continue
        if path.name == "__init__.py":
            continue
        if not path.name.endswith(".py"):
            continue
        relative_path = path.relative_to(app_path)
        route = Route.from_file_path(relative_path)
        routes.append(route)

    return routes


def _app_folder_from_import_path(app_import_path: str) -> Path:
    """Get the app folder from an app import path."""
    app_module = importlib.import_module(app_import_path)
    if app_module.__file__ is None:
        # Should not happen, for type narrowing
        raise ValueError(f"App module {app_module} has no __file__ attribute.")
    app_folder = Path(app_module.__file__).parent
    return app_folder



from typing import List

import typer
from pydantic import BaseModel

from typer_router.from_fs import routes_from_app_import_path
from typer_router.route import Route
from typer_router.to_typer import create_typer_app_from_router


class Router(BaseModel):
    routes: List[Route]
    app_import_path: str

    def to_typer(self, **typer_kwargs) -> typer.Typer:
        return create_typer_app_from_router(self, **typer_kwargs)

    def full_import_path_for(self, route: Route) -> str:
        return f"{self.app_import_path}.{route.import_path}"

    @classmethod
    def from_app_import_path(cls, app_import_path: str) -> "Router":
        routes = routes_from_app_import_path(app_import_path)

        return cls(routes=routes, app_import_path=app_import_path)



from tests.fixtures import APP_1_IMPORT_PATH
from tests.typer_assertions import assert_app1_typer_is_correct
from typer_router import Router


def test_creates_router_from_app_import_path():
    app1_router = Router.from_app_import_path(APP_1_IMPORT_PATH)


def test_typer_has_all_commands(app1_router: Router):
    app1_router = Router.from_app_import_path(APP_1_IMPORT_PATH)
    typer_app = app1_router.to_typer(name="app1")
    assert_app1_typer_is_correct(typer_app)



from tests.typer_assertions import assert_app1_typer_is_correct
from typer_router import Router


def test_creates_typer_from_router(app1_router: Router):
    app1_router.to_typer()


def test_typer_has_all_commands(app1_router: Router):
    typer_app = app1_router.to_typer(name="app1")
    assert_app1_typer_is_correct(typer_app)
